using System;
using System.Collections.Generic;
using System.Text;
using System.Collections;
using gudusoft.gsqlparser;
using gudusoft.gsqlparser.Units;
using System.IO;

namespace JoinRelationAnalyze
{
    enum ClauseType
    {
        where, connectby, startwith, orderby
    }

    public class JoinCondition
    {
        public string lefttable, righttable, leftcolumn, rightcolumn;
        public TCustomSqlStatement sql;

        public override int GetHashCode()
        {
            int hashCode = 0;
            if (lefttable != null)
                hashCode += lefttable.GetHashCode();
            if (righttable != null)
                hashCode += righttable.GetHashCode();
            if (leftcolumn != null)
                hashCode += leftcolumn.GetHashCode();
            if (rightcolumn != null)
                hashCode += rightcolumn.GetHashCode();
            if (sql != null && sql != null)
                hashCode += sql.GetHashCode();
            return hashCode;
        }

        public override bool Equals(object obj)
        {
            if (this == obj)
                return true;
            if (!(obj is JoinCondition))
                return false;

            JoinCondition join = (JoinCondition)obj;

            if (this.leftcolumn != null && !this.leftcolumn.Equals(join.leftcolumn))
                return false;
            if (this.rightcolumn != null && !this.rightcolumn.Equals(join.rightcolumn))
                return false;
            if (this.lefttable != null && !this.lefttable.Equals(join.lefttable))
                return false;
            if (this.righttable != null && !this.righttable.Equals(join.righttable))
                return false;
            if (this.sql != null && !this.sql.Equals(join.sql))
                return false;

            if (join.righttable != null && !join.righttable.Equals(this.righttable))
                return false;
            if (join.lefttable != null && !join.lefttable.Equals(this.lefttable))
                return false;
            if (join.rightcolumn != null && !join.rightcolumn.Equals(this.rightcolumn))
                return false;
            if (join.leftcolumn != null && !join.leftcolumn.Equals(this.leftcolumn))
                return false;
            if (join.sql != null && !join.sql.Equals(this.sql))
                return false;

            return true;
        }
    }

    public class TColumn
    {
        public List<string> tableNames = new List<string>();
        public string columnName;
        public string columnPrex;
        public string columnAlias;

        public string getFullName(string tableName)
        {
            if (tableName != null)
            {
                return tableName + "." + columnName;
            }
            else
            {
                return columnName;
            }
        }

        public string getOrigName()
        {
            if (columnPrex != null)
            {
                return columnPrex + "." + columnName;
            }
            else
            {
                return columnName;
            }
        }

    }

    class TTable
    {
        public string tableName;
        public string prefixName;
        public string tableAlias;
    }

    class joinConditonsInExpr
    {

        private TLzCustomExpression expr;
        private joinRelationAnalyze analysis;
        private TCustomSqlStatement stmt;
        public joinConditonsInExpr(joinRelationAnalyze analysis, TLzCustomExpression expr, TCustomSqlStatement stmt)
        {
            this.stmt = stmt;
            this.analysis = analysis;
            this.expr = expr;
        }

        private bool searchJoinVisitor(TLz_Node pnode, Boolean pIsLeafNode)
        {
            TLzCustomExpression lcexpr = (TLzCustomExpression)pnode;
            if (lcexpr.oper == TLzOpType.Expr_Comparison && "=".Equals(lcexpr.opname.AsText))
            {
                TLzCustomExpression slexpr = (TLzCustomExpression)lcexpr.lexpr;
                TLzCustomExpression srexpr = (TLzCustomExpression)lcexpr.rexpr;

                if (((slexpr.oper == TLzOpType.Expr_Attr) || (slexpr.oper == TLzOpType.Expr_OuterJoin) || (srexpr.oper == TLzOpType.Expr_OuterJoin && slexpr.oper == TLzOpType.Expr_Const))
                 && ((srexpr.oper == TLzOpType.Expr_Attr) || (srexpr.oper == TLzOpType.Expr_OuterJoin) || (slexpr.oper == TLzOpType.Expr_OuterJoin && srexpr.oper == TLzOpType.Expr_Const) 
                 || (slexpr.oper == TLzOpType.Expr_OuterJoin && srexpr.oper == TLzOpType.Expr_Case)))
                {
                    TLz_Attr lattr = null, rattr = null;
                    JoinCondition jr = new JoinCondition();
                    jr.sql = stmt;

                    if (slexpr.oper == TLzOpType.Expr_Attr)
                    {
                        lattr = (TLz_Attr)slexpr.lexpr;
                    }
                    else if (slexpr.oper == TLzOpType.Expr_OuterJoin)
                    {
                        lattr = (TLz_Attr)((TLzCustomExpression)slexpr.lexpr).lexpr;
                    }
                    jr.lefttable = lattr != null ? lattr.Prefix : null;
                    jr.leftcolumn = lattr != null ? lattr.lastname : null;

                    if (srexpr.oper == TLzOpType.Expr_Attr)
                    {
                        rattr = (TLz_Attr)srexpr.lexpr;
                        jr.righttable = rattr != null ? rattr.Prefix : null;
                        jr.rightcolumn = rattr != null ? rattr.lastname : null;
                        analysis.joinRelationSet.Add(jr);
                    }
                    else if (srexpr.oper == TLzOpType.Expr_OuterJoin)
                    {
                        rattr = (TLz_Attr)((TLzCustomExpression)srexpr.lexpr).lexpr;
                        jr.righttable = rattr != null ? rattr.Prefix : null;
                        jr.rightcolumn = rattr != null ? rattr.lastname : null;
                        analysis.joinRelationSet.Add(jr);
                    }
                    else if (srexpr.oper == TLzOpType.Expr_Case)
                    {
                        TLzCaseExpression expr = (TLzCaseExpression)srexpr;
                        TLz_List list = expr.WhenThenList;
                        for (int i = 0; i < list.Count(); i++)
                        {
                            TLzCustomExpression thenexpr = ((TLzCaseWhen)list[i]).ThenExpr;
                            if (thenexpr.oper == TLzOpType.Expr_Attr)
                            {
                                rattr = (TLz_Attr)thenexpr.lexpr;
                            }
                            else if (srexpr.oper == TLzOpType.Expr_OuterJoin)
                            {
                                rattr = (TLz_Attr)((TLzCustomExpression)thenexpr.lexpr).lexpr;
                            }

                            JoinCondition condtion = new JoinCondition();
                            condtion.leftcolumn = jr.leftcolumn;
                            condtion.lefttable = jr.lefttable;
                            condtion.sql = jr.sql;
                            condtion.righttable = rattr != null ? rattr.Prefix : null;
                            condtion.rightcolumn = rattr != null ? rattr.lastname : null;
                            analysis.joinRelationSet.Add(condtion);
                        }
                        if (expr.DefaultExpr is TLzCustomExpression)
                        {
                            TLzCustomExpression elseexpr = expr.DefaultExpr;
                            if (elseexpr.oper == TLzOpType.Expr_Attr)
                            {
                                rattr = (TLz_Attr)elseexpr.lexpr;
                            }
                            else if (srexpr.oper == TLzOpType.Expr_OuterJoin)
                            {
                                rattr = (TLz_Attr)((TLzCustomExpression)elseexpr.lexpr).lexpr;
                            }

                            JoinCondition condtion = new JoinCondition();
                            condtion.leftcolumn = jr.leftcolumn;
                            condtion.lefttable = jr.lefttable;
                            condtion.sql = jr.sql;
                            condtion.righttable = rattr != null ? rattr.Prefix : null;
                            condtion.rightcolumn = rattr != null ? rattr.lastname : null;
                            analysis.joinRelationSet.Add(condtion);
                        }
                    }
                }
            }


            if (lcexpr.oper == TLzOpType.Expr_FuncCall)
            {
                TLz_FuncCall func = (TLz_FuncCall)lcexpr.lexpr;
                if (func.args != null)
                {
                    for (int k = 0; k < func.args.Count(); k++)
                    {
                        if (func.args[k] is TLzCustomExpression)
                        {
                            TLzCustomExpression expr = (TLzCustomExpression)func.args[k];
                            expr.InOrderTraverse(searchJoinVisitor);
                        }
                    }
                }
                if (func._analyticFunction != null)
                {
                    TLz_List list = func._analyticFunction._partitionExprList;
                    searchJoinInList(list, stmt);

                    if (func._analyticFunction._sortclause != null)
                    {
                        list = func._analyticFunction._sortclause.list1;
                        searchJoinInList(list, stmt);

                        list = func._analyticFunction._sortclause.list2;
                        searchJoinInList(list, stmt);
                    }
                }

            }
            else if (lcexpr.oper == TLzOpType.Expr_subquery)
            {
                if (lcexpr.lexpr is TSelectSqlStatement)
                {
                    foreach (TLzField field in ((TSelectSqlStatement)lcexpr.lexpr).Fields)
                    {
                        analysis.searchSubQuery((TSelectSqlStatement)lcexpr.lexpr);
                    }
                }
            }
            else if (lcexpr.oper == TLzOpType.Expr_Case)
            {
                TLzCaseExpression expr = (TLzCaseExpression)lcexpr;
                TLzCustomExpression conditionExpr = expr.ConditionExpr;
                if (conditionExpr != null)
                {
                    conditionExpr.InOrderTraverse(searchJoinVisitor);
                }
                TLzCustomExpression defaultExpr = expr.DefaultExpr;
                if (defaultExpr != null)
                {
                    defaultExpr.InOrderTraverse(searchJoinVisitor);
                }
                TLz_List list = expr.WhenThenList;
                searchJoinInList(list, stmt);
            }
            else if (lcexpr.oper == TLzOpType.Expr_Exists)
            {
                TLzCustomExpression expr = (TLzCustomExpression)lcexpr;
                if (expr.rexpr is TSelectSqlStatement)
                {
                    foreach (TLzField field in ((TSelectSqlStatement)expr.rexpr).Fields)
                    {
                        analysis.searchSubQuery((TSelectSqlStatement)expr.rexpr);
                    }
                }
            }
            return true;
        }

        private void searchJoinInList(TLz_List list, TCustomSqlStatement stmt)
        {
            if (list != null)
            {
                for (int i = 0; i < list.Count(); i++)
                {
                    List<TLzCustomExpression> exprList = new List<TLzCustomExpression>();

                    if (list[i] is TLz_SortGroupBy)
                    {
                        exprList.Add((TLzCustomExpression)((TLz_SortGroupBy)list[i]).node);
                    }
                    else if (list[i] is TLzCustomExpression)
                    {
                        exprList.Add((TLzCustomExpression)list[i]);
                    }
                    else if (list[i] is TLzCaseWhen)
                    {
                        exprList.Add(((TLzCaseWhen)list[i]).WhenExpr);
                        exprList.Add(((TLzCaseWhen)list[i]).ThenExpr);
                    }

                    foreach (TLzCustomExpression lcexpr in exprList)
                    {
                        lcexpr.InOrderTraverse(searchJoinVisitor);
                    }
                }
            }
        }

        public void searchExpression()
        {
            this.expr.InOrderTraverse(searchJoinVisitor);
        }
    }


    public class joinRelationAnalyze
    {
        private static bool isOutputFile;
        private StringBuilder buffer = new StringBuilder();
        private Hashtable cteMap = new Hashtable();
        private Hashtable tableAliasMap = new Hashtable();
        private List<TCustomSqlStatement> searchInSubQuerys = new List<TCustomSqlStatement>();
        private List<TCustomSqlStatement> searchInTables = new List<TCustomSqlStatement>();
        private List<TCustomSqlStatement> searchInClauses = new List<TCustomSqlStatement>();
        public Hashtable queryAliasMap = new Hashtable();
        public HashSet<JoinCondition> joinRelationSet = new HashSet<JoinCondition>();
        private List<JoinCondition> conditions = new List<JoinCondition>();

        public string AnalysisResult
        {
            get { return buffer.ToString(); }
        }

        public List<JoinCondition> JoinConditions
        {
            get { return conditions; }
        }


        static void Main(string[] args)
        {
            if (args.Length == 0)
            {
                Console.WriteLine("Usage: {0} <sql script file path> <output file path>", Environment.GetCommandLineArgs()[0]);
                Console.WriteLine("sql script file path: The sql file will be analyzed.");
                Console.WriteLine("output file path: Option, write the analysis result to the specified file.");
                //Console.Read();
                return;
            }

            string outputFile = null;
            StreamWriter writer = null;
            if (args.Length > 1)
            {
                outputFile = args[1];
                isOutputFile = true;
            }

            if (outputFile != null)
            {
                try
                {
                    writer = new StreamWriter(outputFile);
                    Console.SetOut(writer);
                }
                finally { }
            }
            joinRelationAnalyze analysis = new joinRelationAnalyze(new FileInfo(args[0]), TDbVendor.DbVOracle);
            Console.Write(analysis.AnalysisResult);
            //if (args.Length <= 1)
            //{
            //    Console.Read();
            //}
            //else
            {
                if (writer != null)
                {
                    writer.Close();
                }
            }

        } //main

        public joinRelationAnalyze(string sql, TDbVendor dbVendor)
        {
            TGSqlParser sqlparser = new TGSqlParser(dbVendor);
            sqlparser.SqlText.Text = sql;
            analyzeSQL(sqlparser);
        }

        public joinRelationAnalyze(FileInfo file, TDbVendor dbVendor)
        {
            TGSqlParser sqlparser = new TGSqlParser(dbVendor);
            sqlparser.Sqlfilename = file.ToString();
            analyzeSQL(sqlparser);
        }

        private void analyzeSQL(TGSqlParser sqlparser)
        {
            int ret = sqlparser.Parse();

            if (ret != 0)
            {
                buffer.AppendLine(sqlparser.ErrorMessages);
                return;
            }
            else
            {

                TCustomSqlStatement select = (TCustomSqlStatement)sqlparser.SqlStatements[0];
                if (select.cte != null && select.cte.ctes != null && select.cte.ctes.Count() > 0)
                {
                    for (int i = 0; i < select.cte.ctes.Count(); i++)
                    {
                        TCommonTableExpression expression = (TCommonTableExpression)select.cte.ctes[i];
                        cteMap.Add(expression.ndTablename.AsText, expression.ndSelect_with_parens);
                    }
                }

                analyzeStatement(select);
            }

            buffer.AppendLine("JoinTable1\tJoinColumn1\tJoinTable2\tJoinColumn2");

            conditions.Clear();

            foreach (JoinCondition join in joinRelationSet)
            {
                string lefttable = join.lefttable;
                string righttable = join.righttable;
                string leftcolumn = join.leftcolumn;
                string rightcolumn = join.rightcolumn;

                if ((lefttable == null || lefttable.Length == 0) && (righttable == null || righttable.Length == 0))
                    continue;

                List<string[]> leftJoinNameList = getRealName(lefttable, leftcolumn, join.sql);
                List<string[]> rightJoinNameList = getRealName(righttable, rightcolumn, join.sql);

                foreach (string[] leftJoinNames in leftJoinNameList)
                {
                    foreach (string[] rightJoinNames in rightJoinNameList)
                    {
                        if (leftJoinNames[0] != null && rightJoinNames[0] != null)
                        {
                            JoinCondition condition = new JoinCondition();
                            condition.lefttable = leftJoinNames[0];
                            condition.righttable = rightJoinNames[0];
                            condition.leftcolumn = leftJoinNames[1];
                            condition.rightcolumn = rightJoinNames[1];

                            if (!conditions.Contains(condition))
                            {
                                conditions.Add(condition);
                                buffer.AppendLine(fillString(condition.lefttable) + "\t" + fillString(condition.leftcolumn) + "\t" + fillString(condition.righttable) + "\t" + fillString(condition.rightcolumn));
                            }
                        }
                    }
                }
            }
        }

        private void analyzeStatement(TCustomSqlStatement select)
        {
            if (select is TSelectSqlStatement)
            {
                TSelectSqlStatement stmt = (TSelectSqlStatement)select;

                searchJoinFromStatement(stmt);

                if (stmt.SelectSetType != TSelectSetType.sltNone)
                {
                    analyzeStatement(stmt.LeftStmt);
                    analyzeStatement(stmt.RightStmt);
                }
                else
                {
                    foreach (TLzField field in select.Fields)
                    {
                        searchFields(field, select);
                    }
                }
            }
            else
            {
                foreach (TLzField field in select.Fields)
                {
                    searchFields(field, select);
                }
            }
        }

        private void searchJoinFromStatement(TSelectSqlStatement stmt)
        {
            if (stmt.JoinTables != null)
            {
                for (int i = 0; i < stmt.JoinTables.Count(); i++)
                {
                    TLzJoin join = stmt.JoinTables[i];
                    if (join.JoinItems != null)
                    {
                        for (int j = 0; j < join.JoinItems.Count(); j++)
                        {
                            TLzJoinItem joinItem = join.JoinItems[j];
                            TLzCustomExpression expr = joinItem.JoinQual;
                            searchExpression(expr, stmt);
                        }
                    }
                }
            }
        }

        private List<string[]> getRealName(string tableAlias, string columnAlias, TCustomSqlStatement stmt)
        {
            List<string[]> nameList = new List<string[]>();
            string tableName = null;
            string columnName = columnAlias;
            if ((tableAlias == null || tableAlias.Length == 0) && stmt is TSelectSqlStatement && ((TSelectSqlStatement)stmt).Tables.Count() == 1 && ((TSelectSqlStatement)stmt).Tables[0].AliasClause == null)
            {
                tableName = ((TSelectSqlStatement)stmt).Tables[0].TableName;
                nameList.Add(new string[] { tableName, columnName });
                return nameList;
            }
            else if (tableAlias == null || tableAlias.Length == 0)
            {
                nameList.Add(new string[] { null, columnName });
                return nameList;
            }

            if (tableAliasMap.ContainsKey(tableAlias.ToLower() + ":" + stmt.AsText))
            {
                tableName = (string)tableAliasMap[tableAlias.ToLower() + ":" + stmt.AsText];
                nameList.Add(new string[] { tableName, columnName });
                return nameList;
            }
            else if (tableAliasMap.ContainsKey(tableAlias.ToLower()) && !containsKey(tableAliasMap, tableAlias.ToLower() + ":"))
            {
                tableName = (string)tableAliasMap[tableAlias.ToLower()];
                nameList.Add(new string[] { tableName, columnName });
                return nameList;
            }
            else
            {
                if (queryAliasMap.ContainsKey(tableAlias.ToLower()))
                {
                    Object value = queryAliasMap[tableAlias.ToLower()];
                    if (value is TSelectSqlStatement)
                    {
                        TSelectSqlStatement sql = (TSelectSqlStatement)value;
                        getRealNameFromSql(nameList, columnAlias, stmt, sql);
                    }
                    return nameList;
                }
                else if (stmt is TSelectSqlStatement)
                {
                    findTableByAlias(nameList, (TSelectSqlStatement)stmt, tableAlias, columnAlias, new List<TSelectSqlStatement>());
                    return nameList;
                }
                else return nameList;
            }
        }

        private void getRealNameFromSql(List<string[]> nameList, string columnAlias, TCustomSqlStatement stmt, TSelectSqlStatement sql)
        {
            string tableName = null;
            string columnName = null;

            if (sql.SelectSetType != TSelectSetType.sltNone)
            {
                getRealNameFromSql(nameList, columnAlias, stmt, sql.LeftStmt);
                getRealNameFromSql(nameList, columnAlias, stmt, sql.RightStmt);
            }
            else
            {
                foreach (TLzField field in sql.Fields)
                {
                    switch (field.FieldType)
                    {
                        case TLzFieldType.lftAttr:
                            TColumn column = attrToColumn(field, sql);
                            if (((column.columnAlias == null || column.columnAlias.Length == 0) && columnAlias.Trim().Equals(column.columnName.Trim()))
                                || ((column.columnAlias != null && column.columnAlias.Length > 0) && columnAlias.Trim().Equals(column.columnAlias.Trim())))
                            {
                                if (column.columnPrex != null)
                                {
                                    if (tableAliasMap.ContainsKey(column.columnPrex.ToLower() + ":" + stmt.AsText))
                                    {
                                        tableName = (string)tableAliasMap[column.columnPrex.ToLower() + ":" + stmt.AsText];
                                    }
                                    else if (tableAliasMap.ContainsKey(column.columnPrex.ToLower()))
                                    {
                                        tableName = (string)tableAliasMap[column.columnPrex.ToLower()];
                                    }
                                }
                                else
                                {
                                    TLzTable firstTable = sql.Tables[0];
                                    tableName = firstTable.AsText;
                                }

                                if (column.columnAlias != null)
                                    columnName = column.columnName;
                            }
                            break;
                    }
                }
                nameList.Add(new string[] { tableName, columnName });
            }
        }

        private void findTableByAlias(List<string[]> nameList, TSelectSqlStatement stmt, string tableAlias, string columnAlias, List<TSelectSqlStatement> stats)
        {
            if (stats.Contains(stmt))
                return;
            else stats.Add(stmt);

            if (stmt.SelectSetType != TSelectSetType.sltNone)
            {
                findTableByAlias(nameList, stmt.LeftStmt, tableAlias, columnAlias, stats);
                findTableByAlias(nameList, stmt.RightStmt, tableAlias, columnAlias, stats);
            }
            else
            {
                foreach (TLzTable table in stmt.Tables)
                {
                    if (table.TableAlias != null)
                    {
                        if (table.TableAlias.Equals(tableAlias, StringComparison.OrdinalIgnoreCase))
                        {
                            nameList.Add(new string[] { table.TableName, columnAlias });
                            return;
                        }
                    }
                    else if (table.TableName != null)
                    {
                        if (table.TableName.Equals(tableAlias, StringComparison.OrdinalIgnoreCase))
                        {
                            nameList.Add(new string[] { table.TableName, columnAlias });
                            return;
                        }
                    }
                }
            }
            if (nameList.Count == 0 && stmt.OwnerStmt is TSelectSqlStatement)
            {
                findTableByAlias(nameList, (TSelectSqlStatement)stmt.OwnerStmt, tableAlias, columnAlias, stats);
            }

        }

        private bool containsKey(Hashtable tableAliasMap, string key)
        {
            ICollection collection = tableAliasMap.Keys;
            foreach (string str in collection)
            {
                if (str.StartsWith(key))
                    return true;
            }
            return false;
        }

        private string fillString(string text)
        {
            int tabLength = 8;
            if (isOutputFile)
            {
                tabLength = 9;
            }

            if (text.Length < tabLength)
                text += "\t";
            return text;
        }


        public void searchFields(TLzField field, TCustomSqlStatement select)
        {
            TLzTableList tables = select.Tables;
            switch (field.FieldType)
            {
                case TLzFieldType.lftAttr:
                    searchTables(select);
                    searchClauses(select);
                    break;
                case TLzFieldType.lftExpression:

                    searchExpression(field.FieldExpr, select);
                    searchTables(select);
                    searchClauses(select);

                    if (field.FieldExpr.oper == TLzOpType.Expr_FuncCall)
                    {
                        TLz_FuncCall func = (TLz_FuncCall)field.FieldExpr.lexpr;
                        // buffer.AppendLine("function name {0}", func.funcname.AsText);

                        // check column in function arguments
                        if (func.args != null)
                        {
                            for (int k = 0; k < func.args.Count(); k++)
                            {
                                if (func.args[k] is TLzCustomExpression)
                                {
                                    TLzCustomExpression expr = (TLzCustomExpression)func.args[k];
                                    searchExpression(expr, select);
                                }
                            }
                        }
                        else
                        {
                            if (select.Tables[0].AliasClause != null)
                            {
                                string alias = select.Tables[0].AliasClause.aliastext;
                                if (!tableAliasMap.ContainsKey(alias.ToLower().Trim() + ":" + select.AsText))
                                {
                                    tableAliasMap.Add(alias.ToLower().Trim() + ":" + select.AsText, select.Tables[0].TableName.Trim());
                                }
                                if (!tableAliasMap.ContainsKey(alias.ToLower().Trim()))
                                {
                                    tableAliasMap.Add(alias.ToLower().Trim(), select.Tables[0].TableName.Trim());
                                }
                            }
                        }

                        if (func._analyticFunction != null)
                        {
                            TLz_List list = func._analyticFunction._partitionExprList;

                            searchExpressionList(select, list);

                            if (func._analyticFunction._sortclause != null)
                            {
                                list = func._analyticFunction._sortclause.list1;
                                searchExpressionList(select, list);

                                list = func._analyticFunction._sortclause.list2;
                                searchExpressionList(select, list);

                            }
                        }

                        // check order by clause
                        //if (select is TSelectSqlStatement && ((TSelectSqlStatement)select).GroupbyClause != null)
                        //{
                        //    for (int j = 0; j < ((TSelectSqlStatement)select).GroupbyClause.GroupItems.Count(); j++)
                        //    {
                        //        TLzGroupByItem i = (TLzGroupByItem)((TSelectSqlStatement)select).GroupbyClause.GroupItems[j];
                        //        searchExpression((TLzCustomExpression)i._ndExpr, select);
                        //        searchTables(select);
                        //    }

                        //}

                    }
                    break;
                case TLzFieldType.lftSubquery:
                    if (field.SubQuery is TSelectSqlStatement)
                    {
                        searchSubQuery((TSelectSqlStatement)field.SubQuery);
                    }
                    break;
                default:
                    buffer.AppendLine("searchFields of type: " + field.FieldType + " not implemented yet");
                    break;
            }
        }


        private void searchExpressionList(TCustomSqlStatement select, TLz_List list)
        {
            if (list == null)
                return;

            for (int i = 0; i < list.Count(); i++)
            {
                TLzCustomExpression lcexpr = null;
                if (list[i] is TLz_SortGroupBy)
                {
                    lcexpr = (TLzCustomExpression)((TLz_SortGroupBy)list[i]).node;
                }
                else if (list[i] is TLzCustomExpression)
                {
                    lcexpr = (TLzCustomExpression)list[i];
                }

                if (lcexpr != null)
                {
                    searchExpression(lcexpr, select);
                }
            }
        }

        private void searchClauses(TCustomSqlStatement select)
        {
            if (!searchInClauses.Contains(select))
            {
                searchInClauses.Add(select);
            }
            else
            {
                return;
            }
            if (select is TSelectSqlStatement)
            {

                TSelectSqlStatement statement = (TSelectSqlStatement)select;
                Hashtable clauseTable = new Hashtable();

                //if (statement.SortClause != null)
                //{
                //    TLzOrderByList sortList = (TLzOrderByList)statement.SortClause;
                //    for (int i = 0; i < sortList.Count(); i++)
                //    {
                //        TLzOrderBy orderBy = sortList[i];
                //        TLzCustomExpression expr = orderBy.SortExpr;
                //        clauseTable.Add(expr, ClauseType.orderby);
                //    }
                //}


                if (statement.WhereClause != null)
                {
                    clauseTable.Add((TLzCustomExpression)(statement.WhereClause), ClauseType.where);
                }
                //if (statement.ConnectByClause != null)
                //{
                //    clauseTable.Add((TLzCustomExpression)statement.ConnectByClause, ClauseType.connectby);
                //}
                //if (statement.StartwithClause != null)
                //{
                //    clauseTable.Add((TLzCustomExpression)statement.StartwithClause, ClauseType.startwith);
                //}
                foreach (TLzCustomExpression expr in clauseTable.Keys)
                {
                    ClauseType type = (ClauseType)clauseTable[expr];
                    searchExpression(expr, select);
                    searchTables(select);

                }
            }
        }

        void searchTables(TCustomSqlStatement select)
        {
            if (!searchInTables.Contains(select))
            {
                searchInTables.Add(select);
            }
            else
            {
                return;
            }

            TLzTableList tables = select.Tables;

            if (tables.Count() == 1)
            {
                TLzTable lzTable = tables[0];
                if ((lzTable.TableType == TLzTableType.lttAttr)
                    && (lzTable.TableAlias.Length == 0))
                {
                    if (cteMap.ContainsKey(lzTable.Name))
                    {
                        searchSubQuery((TSelectSqlStatement)cteMap[lzTable.Name]);
                    }
                    else
                    {
                        if (lzTable.AliasClause != null)
                        {
                            string alias = lzTable.AliasClause.aliastext;
                            if (!tableAliasMap.ContainsKey(alias.ToLower().Trim() + ":" + select.AsText))
                            {
                                tableAliasMap.Add(alias.ToLower().Trim() + ":" + select.AsText, lzTable.TableName.Trim());
                            }
                            if (!tableAliasMap.ContainsKey(alias.ToLower().Trim()))
                            {
                                tableAliasMap.Add(alias.ToLower().Trim(), lzTable.TableName.Trim());
                            }
                        }
                    }
                }
            }

            foreach (TLzTable lztable in tables)
            {
                switch (lztable.TableType)
                {
                    case TLzTableType.lttAttr:
                        TTable table = TLzTaleToTable(lztable);
                        string alias = table.tableAlias;
                        if (alias != null)
                            alias = alias.Trim();
                        else if (table.tableName != null)
                            alias = table.tableName.Trim();

                        if (cteMap.ContainsKey(lztable.Name))
                        {
                            searchSubQuery((TSelectSqlStatement)cteMap[lztable.Name]);
                        }
                        else
                        {
                            if (alias != null)
                            {
                                if (!tableAliasMap.ContainsKey(alias.ToLower() + ":" + select.AsText))
                                {
                                    tableAliasMap.Add(alias.ToLower() + ":" + select.AsText, table.tableName.Trim());
                                }
                                if (!tableAliasMap.ContainsKey(alias.ToLower()))
                                {
                                    tableAliasMap.Add(alias.ToLower(), table.tableName.Trim());
                                }
                            }
                        }
                        break;
                    case TLzTableType.lttSubquery:
                        if (lztable.AliasClause != null)
                        {
                            string tableAlias = lztable.AliasClause.aliastext.Trim();
                            if (!queryAliasMap.ContainsKey(tableAlias.ToLower()))
                            {
                                queryAliasMap.Add(tableAlias.ToLower(), (TSelectSqlStatement)lztable.SubQuery);
                            }
                        }
                        searchSubQuery((TSelectSqlStatement)lztable.SubQuery);
                        break;
                    default:
                        break;
                }
            }
        }

        public void searchSubQuery(TSelectSqlStatement select)
        {
            if (!searchInSubQuerys.Contains(select))
            {
                searchInSubQuerys.Add(select);
            }
            else
            {
                return;
            }

            searchJoinFromStatement(select);

            if (select.SelectSetType != TSelectSetType.sltNone)
            {
                searchSubQuery(select.LeftStmt);
                searchSubQuery(select.RightStmt);
            }
            else
            {
                foreach (TLzField field in select.Fields)
                {
                    searchFields(field, select);
                }
            }
        }

        public TColumn attrToColumn(TLzField field, TCustomSqlStatement stmt)
        {
            TColumn column = new TColumn();
            TLz_Attr attr = field.FieldAttr;


            column.columnAlias = field.FieldAlias;
            column.columnName = attr.LastToken.AsText;

            if (attr.AsText.IndexOf(".") > 0)
            {
                column.columnPrex = attr.AsText.Substring(0, attr.AsText.Length - column.columnName.Length - ".".Length);
                if (attr.LastToken.ParentToken != null)
                {
                    if (!column.tableNames.Contains(attr.LastToken.ParentToken.AsText))
                    {
                        column.tableNames.Add(attr.LastToken.ParentToken.AsText);
                    }
                }
            }
            else
            {
                TLzTableList tables = stmt.Tables;
                foreach (TLzTable lztable in tables)
                {
                    TTable table = TLzTaleToTable(lztable);
                    if (!column.tableNames.Contains(table.tableName))
                    {
                        column.tableNames.Add(table.tableName);
                    }
                }
            }

            return column;
        }

        TTable TLzTaleToTable(TLzTable lztable)
        {
            TTable table = new TTable();
            if (lztable.TableAttr != null && lztable.TableAttr.LastToken != null)
            {
                table.tableName = lztable.TableAttr.LastToken.AsText;
                if (lztable.TableAttr.AsText.IndexOf(".") > 0)
                {
                    table.prefixName = lztable.TableAttr.AsText.Substring(0, lztable.TableAttr.AsText.Length - table.tableName.Length - ".".Length);
                }
            }

            if (lztable.AliasClause != null)
            {
                table.tableAlias = lztable.AliasClause.AsText;
            }
            return table;
        }

        void searchExpression(TLzCustomExpression expr, TCustomSqlStatement stmt)
        {
            joinConditonsInExpr c = new joinConditonsInExpr(this, expr, stmt);
            c.searchExpression();
        }

    }
}
